import string
import re
import urlparse
import os
from ftplib import FTP
import time
import socket # socket.gaierror
import getopt, sys

sys.path.append('/usr/lib/mdk-check-update')
import pygspawn

program_name = "mdk-check-update"
version = "0.1.0"

verbose_level = 1

# verbosity levels
verbose_error = 1
verbose_timer = 2
verbose_actions = 3
verbose_media = 4
verbose_url = 5
verbose_date = 6

update_interval  = 10*60 # second(s)
first_update = 1 # second(s)

icon_path = "/usr/share/pixmaps/" + program_name + "/"
icon_uptodate = icon_path + "uptodate.png"
icon_obsolete = icon_path + "obsolete.png"
media = []

re_begin_medium = re.compile("^((?:\S|\\ )+)\s+(\S*)\s+\{$")
re_hdlist = re.compile("^\s*hdlist:\s*(.*)$")
re_with_hdlist = re.compile("^\s*with_hdlist:\s*(.*)$")
re_synthesis = re.compile("^\s*synthesis$")
re_end_medium = re.compile("\}$")
re_simplify_url = re.compile("/[^/]+/../") # to find and remove /foo/../
re_urlparse = re.compile("(?:([^:]+)://|)(.*)") # to get scheme and location : (rsync)://(foo)
re_netloc_path = re.compile("([^/]+)(.*)")

# locale.setlocale(locale.LC_ALL,'C')
# months=[]
# for m in range(1,13):
#	months.append(time.strftime("%b",time.strptime(str(m),"%m")))
months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

def usage():
	print "usage: " + os.path.basename(sys.argv[0]) + " [-h,--help] [-v verbose_level]"
	sys.exit(2)

def parse_opts():
	global verbose_level
	try:
		opts, args = getopt.getopt(sys.argv[1:], "hv:", ["help"])
	except getopt.GetoptError:
		# print help information and exit:
		usage()
		sys.exit(2)
	for o, a in opts:
		if o == "-v":
			verbose_level = a
		if o in ("-h", "--help"):
			usage()
			sys.exit()

def reduce_url(url):
	while re_simplify_url.search(url):
		url = re_simplify_url.sub("/", url)
	return url

def get_netloc_path(url):
	netloc_path = re_netloc_path.match(url)
	if netloc_path:
		return netloc_path.groups() # returns (netloc, path)

def get_type_url(url):
	# split url
	url_parsed = re_urlparse.match(url)
	if url_parsed:
		return url_parsed.groups() # returns the tuple (type, url_without_type)
	else:
		if verbose_level >= verbose_error:
			print "Error while parsing url"

def read_config():
	global media
	if verbose_level >= verbose_actions:
		print "Reading config ..."
	media = []
	try:
		f = open("/etc/urpmi/urpmi.cfg","r")
		l = f.readlines()
		f.close()
	except IOError:
		if verbose_level >= verbose_error:
			print "Error while reading urpmi config file."
	else:
		for line in l:
			begin_medium = re_begin_medium.match(line)
			hdlist = re_hdlist.match(line)
			with_hdlist = re_with_hdlist.match(line)
			synthesis = re_synthesis.match(line)
			end_medium = re_end_medium.match(line)
			if begin_medium:
				medium_info = {}
				medium_info['name'] = begin_medium.group(1)
				medium_info['type'], medium_info['url'] = get_type_url(begin_medium.group(2))
				if verbose_level >= verbose_media:
					print "Media: " + medium_info['name']
					if verbose_level >= verbose_url:
						print "Type: " + str(medium_info['type']) + ", url: " + medium_info['url']
			elif hdlist:
				medium_info['hdlist'] = hdlist.group(1)
			elif with_hdlist:
				medium_info['with_hdlist'] = with_hdlist.group(1)
			elif synthesis:
				medium_info['synthesis'] = 1
			elif end_medium:
				media.append(medium_info)

	# try to find additional urls in ~/.mdk-check-update/urls
	additional_urls = {}
	filename = os.getenv("HOME") + "/." + program_name + "/urls"
	if verbose_level >= verbose_actions:
		print "Searching additional urls in " + filename + " ..."
	try:
		f = open(filename)
		l = f.readlines()
		f.close()
	except IOError:
		if verbose_level >= verbose_actions:
			print "Error while reading additional urls file."
	else:
		# fill additonal_urls with medium_name as key and url as value
		for line in l:
			try:
				medium_name, url = line.split()
				additional_urls[medium_name] = url
			except ValueError:
				if verbose_level >= verbose_error:
					print "Invalid file format."
				continue
		# add additional urls in media list
		for medium in media:
			if medium['name'] in additional_urls:
				medium['type'], medium['url'] = get_type_url(additional_urls[medium['name']])
				if verbose_level >= verbose_url:
					print "New url for medium " + medium['name'] + ": (" + medium['type'] + ") " + medium['url']

	if verbose_level >= verbose_actions:
		print "Config read."

ftp_reply = ""
def ftp_reply_keep_last_cb(x):
	global ftp_reply
	ftp_reply = x

def check_media():
	if verbose_level >= verbose_actions:
		print "Checking for media updates ..."
	for medium in media:
		if verbose_level >= verbose_media:
			print "Media: " + medium['name']
		# get name and date of local medium hdlist
		local_name = "/var/lib/urpmi/"
		if 'synthesis' in medium:
			local_name += "synthesis."
		local_name += medium['hdlist']
		try:
			local_date =  os.path.getmtime(local_name)
		except OSError:
			if verbose_level >= verbose_error:
				print "Local hdlist not found: " + local_name
			medium['error'] = "Local hdlist not found"
			continue

		if verbose_level >= verbose_url:
			print "Local file: " + local_name
		if verbose_level >= verbose_date:
			print "Local file date: " + str(local_date)

		# get name of remote hdlist
		if 'with_hdlist' in medium:
			hdlist = medium['with_hdlist']
		else:
			hdlist = medium['hdlist']

		# get complete name and date of remote hdlist and adjust local date if needed
		remote_name = ""
		remote_name = reduce_url(medium['url'] + "/" + hdlist)
		if verbose_level >= verbose_url:
			print "Remote file: " + remote_name
		remote_date = 0
		if medium['type'] == 'ftp':
			global ftp_reply
			netloc, path = get_netloc_path(remote_name)
			try:
				ftp = FTP(netloc)
				# anonymous login with a plausible e-mail address
				ftp.login("anonymous", "anonymous@" + os.getenv("HOSTNAME"))
				ftp_reply = ""
				ftp.retrlines('LIST ' + path, ftp_reply_keep_last_cb)
				ftp.close()
			#except (socket.gaierror, socket.error):
			# what about ftplib.error_perm ?
			except: # errm, quite bad :-)
				if verbose_level >= verbose_error:
					print "Unable to find hdlist on ftp server."
				medium['error'] = "Remote hdlist (ftp) not found."
				continue

			# keep the date
			ftp_reply_date = ftp_reply.split()[5:8]
			if len(ftp_reply_date) == 3:
				(month, day, other) = ftp_reply_date
			else:
				if verbose_level >= verbose_error:
					print "File not found (ftp)."
				medium['error'] = "File not found (ftp)."
				continue				
			# convert the month name in number
			if month in months:
				month = months.index(month) + 1
			else:
				if verbose_level >= verbose_error:
					print "Bad date (month) format: " + month
				medium['error'] = "Bad date format (ftp)."
				continue
			# try to guess date format
			if ':' in other:
				# other is the date
				date_format = "%m %d %H:%M %Y"
				ftp_date = string.join([str(month), day, other, str(time.gmtime().tm_year)])
			else:
				# other is the year
				date_format = "%m %d %Y"
				ftp_date = string.join([str(month), day, other])
			if verbose_level >= verbose_date:
				print "Remote ftp file date: " + ftp_date

			# convert to unix date
			try:
				remote_date = time.mktime(time.strptime(ftp_date, date_format))
				remote_date = time.mktime(time.strptime(ftp_date, date_format))				
			       	# very very ugly hack not to care about seconds or other things that don't matter
				local_date = time.mktime(time.strptime(time.strftime(date_format, time.localtime(local_date)), date_format))
			except ValueError:
				if verbose_level >= verbose_error:
					print "Bad date format: " + ftp_date + ", should be: " + date_format
				medium['error'] = "Bad date format (ftp)."
				continue
		elif medium['type'] == 'rsync':
			# use rsync to find the date of remote hdlist
			try:
				out, err, exit_code = pygspawn.g_spawn_command_line_sync("/usr/bin/rsync " + remote_name)
			except IOError:
				if verbose_level >= verbose_error:
					print "Unable to find hdlist on rsync server."
				medium['error'] = "Remote file (rsync) not found."
				continue
			l = out.split()
			# if stdout is empty, skip this medium
			if not len(l) or exit_code :
				if verbose_level >= verbose_error:
					print "Error in rsync:\n" + err
					print "rsync exit code: " + str(exit_code)
				medium['error'] = "Remote file (rsync) not found."
				continue
			date = l[-3]
			hour = l[-2]
			if verbose_level >= verbose_date:
				print "Remote rsync file date: " + string.join([date, hour])
			try:
				remote_date = time.mktime(time.strptime(string.join([date, hour]), "%Y/%m/%d %H:%M:%S"))
			except ValueError:
				if verbose_level >= verbose_error:
					print "Bad date format: " + str(date) + ", " + str(hour)
				medium['error'] = "Bad date format (rsync)."
				continue

		if verbose_level >= verbose_date:
			print "Remote file date: " + str(remote_date)

		if (remote_date > local_date):
			# a new hdlist is available for this medium
			medium['update_available'] = 1

	if verbose_level >= verbose_actions:
		print "Media updates checking done."

def is_obsolete(medium):
	return 'update_available' in medium

def get_medium_name(medium):
	if 'name' in medium:
		return medium['name']
	else:
		return ''

def list_obsolete_media():
	obsolete_media = filter(is_obsolete, media)
	obsolete_media_names = map(get_medium_name, obsolete_media)
	return string.join(obsolete_media_names, ' ')
